# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Tuple, Union, Optional, List
import logging
import asyncio
import hashlib
import hmac

from nio import (AsyncClient, Event as NioEvent, GroupEncryptionError, LoginError,
                 MatrixRoom as NioRoom, RoomMemberEvent as NioMemberEvent, MatrixUser as NioUser,
                 AsyncClientConfig)

from mautrix.types import (Filter, RoomFilter, EventFilter, RoomEventFilter, StateFilter,
                           EventType, RoomID, Serializable, JSON, MessageEvent, Event, UserID,
                           EncryptedEvent, StateEvent, Membership)
from mautrix.bridge.db import UserProfile
from mautrix.bridge.db.nio_state_store import NioStore, DBAccount


class EncryptionManager:
    loop: asyncio.AbstractEventLoop
    log: logging.Logger = logging.getLogger("mau.e2ee")
    client: AsyncClient

    bot_mxid: UserID
    login_shared_secret: bytes
    _id_prefix: str
    _id_suffix: str

    sync_task: asyncio.Task

    def __init__(self, bot_mxid: UserID, login_shared_secret: str, homeserver_address: str,
                 user_id_prefix: str, user_id_suffix: str, device_name: str,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        self.loop = loop or asyncio.get_event_loop()
        self.bot_mxid = bot_mxid
        self.device_name = device_name
        self._id_prefix = user_id_prefix
        self._id_suffix = user_id_suffix
        self.login_shared_secret = login_shared_secret.encode("utf-8")
        config = AsyncClientConfig(store=NioStore, encryption_enabled=True,
                                   pickle_key="mautrix-python", store_sync_tokens=True)
        device_id = DBAccount.first_device_id(self.bot_mxid)
        if device_id:
            self.log.debug(f"Found device ID in database: {device_id}")
        self.client = AsyncClient(homeserver=homeserver_address, user=bot_mxid,
                                  device_id=device_id, config=config, store_path="3:<")

    def _init_load_profiles(self) -> None:
        self.log.debug("Loading room and member list into encryption client")
        for profile in UserProfile.all_except(self._id_prefix, self._id_suffix, self.bot_mxid):
            try:
                room = self.client.rooms[profile.room_id]
            except KeyError:
                room = self.client.rooms[profile.room_id] = NioRoom(profile.room_id, self.bot_mxid,
                                                                    encrypted=True)
            user = NioUser(profile.user_id, profile.displayname, profile.avatar_url)
            if profile.membership == Membership.JOIN:
                room.users[profile.user_id] = user
            elif profile.membership == Membership.INVITE:
                room.invited_users[profile.user_id] = user

    def _ignore_user(self, user_id: str) -> bool:
        return (user_id.startswith(self._id_prefix) and user_id.endswith(self._id_suffix)
                and user_id != self.bot_mxid)

    async def handle_room_membership(self, evt: StateEvent) -> None:
        if self._ignore_user(evt.state_key):
            return
        try:
            room = self.client.rooms[evt.room_id]
        except KeyError:
            room = self.client.rooms[evt.room_id] = NioRoom(evt.room_id, self.bot_mxid,
                                                            encrypted=True)
            await self.client.joined_members(evt.room_id)
        nio_evt = NioMemberEvent.from_dict(evt.serialize())
        try:
            if room.handle_membership(nio_evt):
                self.client._invalidate_session_for_member_event(evt.room_id)
        except Exception:
            self.log.exception("matrix-nio failed to handle membership event")

    async def handle_room_encryption(self, evt: StateEvent) -> None:
        try:
            room = self.client.rooms[evt.room_id]
            room.encrypted = True
        except KeyError:
            self.client.rooms[evt.room_id] = NioRoom(evt.room_id, self.bot_mxid, encrypted=True)
            await self.client.joined_members(evt.room_id)

    async def add_room(self, room_id: RoomID, members: Optional[List[UserID]] = None,
                       encrypted: bool = False) -> None:
        if room_id in self.client.invited_rooms:
            del self.client.invited_rooms[room_id]
        try:
            room = self.client.rooms[room_id]
            room.encrypted = encrypted
        except KeyError:
            room = self.client.rooms[room_id] = NioRoom(room_id, self.bot_mxid, encrypted=True)
        if members:
            update = False
            for member in members:
                if not self._ignore_user(member):
                    update = room.add_member(member, "", "") or update
            if update:
                self.client._invalidate_session_for_member_event(room_id)
        else:
            await self.client.joined_members(room_id)

    async def encrypt(self, room_id: RoomID, event_type: EventType,
                      content: Union[Serializable, JSON]) -> Tuple[EventType, JSON]:
        serialized = content.serialize() if isinstance(content, Serializable) else content
        type_str = str(event_type)
        retries = 0
        while True:
            try:
                type_str, encrypted = self.client.encrypt(room_id, type_str, serialized)
                break
            except GroupEncryptionError:
                if retries > 3:
                    self.log.error("Got GroupEncryptionError again, giving up")
                    raise
                retries += 1
                self.log.debug("Got GroupEncryptionError, sharing group session and trying again")
                await self.client.share_group_session(room_id, ignore_unverified_devices=True)
        event_type = EventType.find(type_str)
        try:
            encrypted["m.relates_to"] = serialized["m.relates_to"]
        except KeyError:
            pass
        return event_type, encrypted

    def decrypt(self, event: EncryptedEvent) -> MessageEvent:
        serialized = event.serialize()
        event = self.client.decrypt_event(NioEvent.parse_encrypted_event(serialized))
        try:
            event.source["content"]["m.relates_to"] = serialized["content"]["m.relates_to"]
        except KeyError:
            pass
        return Event.deserialize(event.source)

    async def start(self) -> None:
        self.log.debug("Logging in with bridge bot user")
        password = hmac.new(self.login_shared_secret, self.bot_mxid.encode("utf-8"),
                            hashlib.sha512).hexdigest()
        resp = await self.client.login(password, device_name=self.device_name)
        if isinstance(resp, LoginError):
            raise Exception(f"Failed to log in with bridge bot: {resp}")
        self._init_load_profiles()
        self.sync_task = self.loop.create_task(self.client.sync_forever(
            timeout=30000, sync_filter=self._filter.serialize()))
        self.log.info("End-to-bridge encryption support is enabled")

    def stop(self) -> None:
        self.sync_task.cancel()

    @property
    def _filter(self) -> Filter:
        all_events = EventType.find("*")
        return Filter(
            account_data=EventFilter(types=[all_events]),
            presence=EventFilter(not_types=[all_events]),
            room=RoomFilter(
                include_leave=False,
                state=StateFilter(not_types=[all_events]),
                timeline=RoomEventFilter(not_types=[all_events]),
                account_data=RoomEventFilter(not_types=[all_events]),
                ephemeral=RoomEventFilter(not_types=[all_events]),
            ),
        )



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, List, Set, Iterable, Dict, Any
from datetime import datetime

from sqlalchemy import Column, PickleType, LargeBinary, String, Boolean, DateTime, Text, select
from sqlalchemy.engine.base import Engine
from nio.crypto import (OlmAccount, SessionStore, GroupSessionStore, DeviceStore, Session,
                        InboundGroupSession, OlmDevice, OutgoingKeyRequest)

from mautrix.util.db import Base
from mautrix.types import UserID, RoomID, SyncToken


class DBAccount(Base):
    __tablename__ = "nio_account"

    user_id: UserID = Column(String(255), primary_key=True)
    device_id: str = Column(String(255), primary_key=True)
    shared: bool = Column(Boolean, nullable=False)
    sync_token: SyncToken = Column(Text, nullable=False)
    account: bytes = Column(LargeBinary, nullable=False)

    @classmethod
    def first_device_id(cls, user_id: UserID) -> Optional[str]:
        acc = cls._select_one_or_none(cls.c.user_id == user_id)
        if not acc:
            return None
        return acc.device_id

    @classmethod
    def get(cls, user_id: UserID, device_id: str) -> Optional['DBAccount']:
        return cls._select_one_or_none((cls.c.user_id == user_id) & (cls.c.device_id == device_id))

    @classmethod
    def set_sync_token(cls, sync_token: SyncToken) -> None:
        with cls.db.begin() as conn:
            conn.execute(cls.t.update().values(sync_token=sync_token))

    @classmethod
    def get_sync_token(cls) -> Optional[SyncToken]:
        rows = cls.db.execute(select([cls.c.sync_token]))
        try:
            return SyncToken(next(rows)[0])
        except (StopIteration, IndexError):
            return None


class DBOlmSession(Base):
    __tablename__ = "nio_olm_session"

    session_id: str = Column(String(255), primary_key=True)
    sender_key: str = Column(String(255), nullable=False)
    session: bytes = Column(LargeBinary, nullable=False)
    created_at: datetime = Column(DateTime, nullable=False)
    last_used: datetime = Column(DateTime, nullable=False)

    @classmethod
    def all(cls) -> Iterable['DBOlmSession']:
        return cls._select_all()


class DBDeviceKey(Base):
    __tablename__ = "nio_device_key"

    user_id: UserID = Column(String(255), primary_key=True)
    device_id: str = Column(String(255), primary_key=True)
    display_name: str = Column(String(255), nullable=False)
    deleted: bool = Column(Boolean, nullable=False)
    keys: Dict[str, str] = Column(PickleType, nullable=False)

    @classmethod
    def all(cls) -> Iterable['DBDeviceKey']:
        return cls._select_all()

    @classmethod
    def insert_many(cls, data: List[Dict[str, Any]]) -> None:
        cls.db.execute(cls.t.insert(), data)


class DBMegolmInboundSession(Base):
    __tablename__ = "nio_megolm_inbound_session"

    session_id: str = Column(String(255), primary_key=True)
    sender_key: str = Column(String(255), nullable=False)
    fp_key: str = Column(String(255), nullable=False)
    room_id: RoomID = Column(String(255), nullable=False)
    session: bytes = Column(LargeBinary, nullable=False)
    forwarded_chains: List[str] = Column(PickleType, nullable=False)

    @classmethod
    def all(cls) -> Iterable['DBMegolmInboundSession']:
        return cls._select_all()


class DBOutgoingKeyRequest(Base):
    __tablename__ = "nio_outgoing_key_request"

    request_id: str = Column(String(255), primary_key=True)
    session_id: str = Column(String(255), nullable=False)
    room_id: RoomID = Column(String(255), nullable=False)
    algorithm: str = Column(String(255), nullable=False)

    @classmethod
    def all(cls) -> Iterable['DBOutgoingKeyRequest']:
        return cls._select_all()

    @classmethod
    def delete_by_id(cls, request_id: str) -> None:
        with cls.db.begin() as conn:
            conn.execute(cls.t.delete().where(cls.c.request_id == request_id))


def init(db_engine: Engine) -> None:
    for table in (DBAccount, DBOlmSession, DBDeviceKey, DBMegolmInboundSession,
                  DBOutgoingKeyRequest):
        table.db = db_engine
        table.t = table.__table__
        table.c = table.t.c
        table.column_names = table.c.keys()


class NioStore:
    user_id: UserID
    device_id: str
    pickle_key: str

    def __init__(self, user_id: UserID, device_id: str, store_path: str, pickle_key: str,
                 store_name: str) -> None:
        self.user_id = user_id
        self.device_id = device_id
        self.pickle_key = pickle_key

    def load_account(self) -> Optional[OlmAccount]:
        """Load the Olm account from the database.

        Returns:
            ``OlmAccount`` object, or ``None`` if it wasn't found for the
                current device_id.

        """
        account = DBAccount.get(self.user_id, self.device_id)
        if not account:
            return None

        return OlmAccount.from_pickle(account.account, self.pickle_key, account.shared)

    def save_account(self, account: OlmAccount) -> None:
        """Save the provided Olm account to the database.

        Args:
            account (OlmAccount): The olm account that will be pickled and
                saved in the database.
        """
        DBAccount(user_id=self.user_id, device_id=self.device_id, shared=account.shared,
                  account=account.pickle(self.pickle_key), sync_token=SyncToken("")).upsert()

    def load_sessions(self) -> SessionStore:
        """Load all Olm sessions from the database.

        Returns:
            ``SessionStore`` object, containing all the loaded sessions.

        """
        session_store = SessionStore()
        for session in DBOlmSession.all():
            session_store.add(session.sender_key, Session.from_pickle(session.session,
                                                                      session.created_at,
                                                                      self.pickle_key))
        return session_store

    def save_session(self, curve_key: str, session: Session) -> None:
        """Save the provided Olm session to the database.

        Args:
            curve_key (str): The curve key that owns the Olm session.
            session (Session): The Olm session that will be pickled and
                saved in the database.
        """
        DBOlmSession(session_id=session.id, sender_key=curve_key, created_at=session.creation_time,
                     last_used=session.use_time, session=session.pickle(self.pickle_key)).upsert()

    def load_inbound_group_sessions(self) -> GroupSessionStore:
        """Load all Olm sessions from the database.

        Returns:
            ``GroupSessionStore`` object, containing all the loaded sessions.

        """
        store = GroupSessionStore()

        for session in DBMegolmInboundSession.all():
            store.add(InboundGroupSession.from_pickle(session.session, session.fp_key,
                                                      session.sender_key, session.room_id,
                                                      self.pickle_key, session.forwarded_chains))

        return store

    def save_inbound_group_session(self, session: InboundGroupSession) -> None:
        """Save the provided Megolm inbound group session to the database.

        Args:
            session (InboundGroupSession): The session to save.
        """
        DBMegolmInboundSession(session_id=session.id, sender_key=session.sender_key,
                               fp_key=session.ed25519, room_id=RoomID(session.room_id),
                               session=session.pickle(self.pickle_key),
                               forwarded_chains=session.forwarding_chain).upsert()

    @staticmethod
    def load_device_keys() -> DeviceStore:
        """Load all the device keys from the database.

        Returns DeviceStore containing the OlmDevices with the device keys.
        """
        store = DeviceStore()
        for device_key in DBDeviceKey.all():
            store.add(OlmDevice(device_key.user_id, device_key.device_id, device_key.keys,
                                display_name=device_key.display_name, deleted=device_key.deleted))
        return store

    @staticmethod
    def save_device_keys(device_keys: Dict[str, Dict[str, OlmDevice]]) -> None:
        """Save the provided device keys to the database.

        Args:
            device_keys: A dictionary
                containing a mapping from a user id to a dictionary containing
                a mapping of a device id to a OlmDevice.
        """
        for user_id, devices_dict in device_keys.items():
            for device_id, device in devices_dict.items():
                DBDeviceKey(user_id=UserID(user_id), device_id=device_id, keys=device.keys,
                            display_name=device.display_name, deleted=device.deleted).upsert()

    @staticmethod
    def load_encrypted_rooms() -> Set[str]:
        """Load the set of encrypted rooms for this account.

        Returns:
            ``Set`` containing room ids of encrypted rooms.

        """
        return set()

    @staticmethod
    def load_outgoing_key_requests() -> Dict[str, OutgoingKeyRequest]:
        """Load the set of outgoing key requests for this account.

        Returns:
            ``Set`` containing request ids of key requests.

        """
        return {request.request_id: OutgoingKeyRequest.from_database(request)
                for request in DBOutgoingKeyRequest.all()}

    @staticmethod
    def add_outgoing_key_request(key_request: OutgoingKeyRequest) -> None:
        """Add an outgoing key request to the store."""
        DBOutgoingKeyRequest(
            request_id=key_request.request_id,
            session_id=key_request.session_id,
            room_id=key_request.room_id,
            algorithm=key_request.algorithm,
        ).upsert()

    @staticmethod
    def remove_outgoing_key_request(key_request: OutgoingKeyRequest) -> None:
        """Remove an active outgoing key request from the store."""
        DBOutgoingKeyRequest.delete_by_id(key_request.request_id)

    def save_encrypted_rooms(self, rooms: Set[str]) -> None:
        """Save the set of room ids for this account."""
        pass

    def delete_encrypted_room(self, room: str) -> None:
        """Delete the a encrypted room from the store."""
        pass

    @staticmethod
    def save_sync_token(token: SyncToken) -> None:
        """Save the given token"""
        DBAccount.set_sync_token(token)

    @staticmethod
    def load_sync_token() -> Optional[SyncToken]:
        return DBAccount.get_sync_token()

    def blacklist_device(self, device: OlmDevice) -> bool:
        """Mark a device as blacklisted.

        Args:
            device (OlmDevice): The device that will be marked as blacklisted

        Returns True if the device was blacklisted, False otherwise, e.g. if
        the device was already blacklisted.

        """
        return False

    def unblacklist_device(self, device: OlmDevice) -> bool:
        """Unmark a device as blacklisted.

        Args:
            device (OlmDevice): The device that will be unmarked as blacklisted

        """
        return False

    def verify_device(self, device: OlmDevice) -> bool:
        """Mark a device as verified.

        Args:
            device (OlmDevice): The device that will be marked as verified

        Returns True if the device was verified, False otherwise, e.g. if the
        device was already verified.

        """
        return False

    def is_device_verified(self, device: OlmDevice) -> bool:
        """Check if a device is verified.

        Args:
            device (OlmDevice): The device that will be checked if it's
                verified.
        """
        return True

    def is_device_blacklisted(self, device: OlmDevice) -> bool:
        """Check if a device is blacklisted.

        Args:
            device (OlmDevice): The device that will be checked if it's
                blacklisted.
        """
        return False

    def unverify_device(self, device: OlmDevice) -> bool:
        """Unmark a device as verified.

        Args:
            device (OlmDevice): The device that will be unmarked as verified

        Returns True if the device was unverified, False otherwise, e.g. if the
        device wasn't verified.

        """
        return False

    def ignore_device(self, device: OlmDevice) -> bool:
        """Mark a device as ignored.

        Args:
            device (OlmDevice): The device that will be marked as blacklisted

        Returns True if the device was ignored, False otherwise, e.g. if
        the device was already ignored.
        """
        return False

    def unignore_device(self, device: OlmDevice) -> bool:
        """Unmark a device as ignored.

        Args:
            device (OlmDevice): The device that will be marked as blacklisted

        Returns True if the device was unignored, False otherwise, e.g. if the
        device wasn't ignored in the first place.
        """
        return False

    def ignore_devices(self, devices: List[OlmDevice]) -> None:
        """Mark a list of devices as ignored.

        This is a more efficient way to mark multiple devices as ignored.

        Args:
            device (list[OlmDevice]): A list of OlmDevices that will be marked
                as ignored.

        """
        pass

    def is_device_ignored(self, device: OlmDevice) -> bool:
        """Check if a device is ignored.

        Args:
            device (OlmDevice): The device that will be checked if it's
                ignored.
        """
        return True



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Iterable

from sqlalchemy import Column, String, Enum

from mautrix.types import RoomID, UserID, ContentURI, Member, Membership
from mautrix.util.db import Base


class UserProfile(Base):
    __tablename__ = "mx_user_profile"

    room_id: RoomID = Column(String(255), primary_key=True)
    user_id: UserID = Column(String(255), primary_key=True)
    membership: Membership = Column(Enum(Membership), nullable=False, default=Membership.LEAVE)
    displayname: str = Column(String, nullable=True)
    avatar_url: ContentURI = Column(String(255), nullable=True)

    def member(self) -> Member:
        return Member(membership=self.membership, displayname=self.displayname,
                      avatar_url=self.avatar_url)

    @classmethod
    def get(cls, room_id: RoomID, user_id: UserID) -> Optional['UserProfile']:
        return cls._select_one_or_none((cls.c.room_id == room_id) & (cls.c.user_id == user_id))

    @classmethod
    def all_except(cls, prefix: str, suffix: str, bot: str) -> Iterable['UserProfile']:
        return cls._select_all(~(cls.c.user_id.startswith(prefix, autoescape=True)
                                 & cls.c.user_id.endswith(suffix, autoescape=True)
                                 & (cls.c.user_id != bot))
                               & ((cls.c.membership == Membership.JOIN)
                                  | (cls.c.membership == Membership.INVITE)))

    @classmethod
    def delete_all(cls, room_id: RoomID) -> None:
        with cls.db.begin() as conn:
            conn.execute(cls.t.delete().where(cls.c.room_id == room_id))



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABC, abstractmethod
import asyncio

from .api.types import EventType, StateEvent, Membership
from .client import Client, InternalEventType, SyncStream


class Dispatcher(ABC):
    client: Client

    def __init__(self, client: Client) -> None:
        self.client = client

    @abstractmethod
    def register(self) -> None:
        pass

    @abstractmethod
    def unregister(self) -> None:
        pass

    def _dispatch(self, event_type: InternalEventType, evt: StateEvent) -> None:
        asyncio.ensure_future(self.client.dispatch_manual_event(event_type, evt),
                              loop=self.client.loop)


class MembershipEventDispatcher(Dispatcher):
    def register(self) -> None:
        self.client.add_event_handler(EventType.ROOM_MEMBER, self.handle)

    def unregister(self) -> None:
        self.client.remove_event_handler(EventType.ROOM_MEMBER, self.handle)

    async def handle(self, evt: StateEvent) -> None:
        if evt.type != EventType.ROOM_MEMBER:
            return

        if evt.content.membership == Membership.JOIN:
            if evt.prev_content.membership != Membership.JOIN:
                self._dispatch(InternalEventType.JOIN, evt)
            else:
                self._dispatch(InternalEventType.PROFILE_CHANGE, evt)
        elif evt.content.membership == Membership.INVITE:
            self._dispatch(InternalEventType.INVITE, evt)
        elif evt.content.membership == Membership.LEAVE:
            if evt.prev_content.membership == Membership.BAN:
                self._dispatch(InternalEventType.UNBAN, evt)
            elif evt.prev_content.membership == Membership.INVITE:
                if evt.state_key == evt.sender:
                    self._dispatch(InternalEventType.REJECT_INVITE, evt)
                else:
                    self._dispatch(InternalEventType.DISINVITE, evt)
            elif evt.state_key == evt.sender:
                self._dispatch(InternalEventType.LEAVE, evt)
            else:
                self._dispatch(InternalEventType.KICK, evt)
        elif evt.content.membership == Membership.BAN:
            self._dispatch(InternalEventType.BAN, evt)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Type
from itertools import chain
import sys

import sqlalchemy as sql
from sqlalchemy.engine.base import Engine

from mautrix.appservice import AppService
from ..util.db import Base
from ..util.program import Program
from .db import SQLStateStore
from .config import BaseBridgeConfig
from .matrix import BaseMatrixHandler

try:
    import uvloop
except ImportError:
    uvloop = None


class Bridge(Program):
    db: Engine
    az: AppService
    state_store_class: Type[SQLStateStore] = SQLStateStore
    state_store: SQLStateStore
    config_class: Type[BaseBridgeConfig]
    config: BaseBridgeConfig
    matrix_class: Type[BaseMatrixHandler]
    matrix: BaseMatrixHandler
    repo_url: str
    markdown_version: str
    real_user_content_key: Optional[str] = None

    def __init__(self, module: str = None, name: str = None, description: str = None,
                 command: str = None, version: str = None,
                 real_user_content_key: Optional[str] = None,
                 config_class: Type[BaseBridgeConfig] = None,
                 matrix_class: Type[BaseMatrixHandler] = None,
                 state_store_class: Type[SQLStateStore] = None) -> None:
        super().__init__(module, name, description, command, version, config_class)
        if real_user_content_key:
            self.real_user_content_key = real_user_content_key
        if matrix_class:
            self.matrix_class = matrix_class
        if state_store_class:
            self.state_store_class = state_store_class

    def prepare_arg_parser(self) -> None:
        super().prepare_arg_parser()
        self.parser.add_argument("-g", "--generate-registration", action="store_true",
                                 help="generate registration and quit")
        self.parser.add_argument("-r", "--registration", type=str, default="registration.yaml",
                                 metavar="<path>",
                                 help="the path to save the generated registration to (not needed "
                                      "for running the bridge)")

    def preinit(self) -> None:
        super().preinit()
        if self.args.generate_registration:
            self.generate_registration()
            sys.exit(0)

    def prepare(self) -> None:
        super().prepare()
        self.prepare_appservice()
        self.prepare_db()
        self.prepare_bridge()

    def prepare_config(self) -> None:
        self.config = self.config_class(self.args.config, self.args.registration,
                                        self.args.base_config)
        if self.args.generate_registration:
            self.config._check_tokens = False
        self.load_and_update_config()

    def generate_registration(self) -> None:
        self.config.generate_registration()
        self.config.save()
        print(f"Registration generated and saved to {self.config.registration_path}")

    def prepare_appservice(self) -> None:
        self.state_store = self.state_store_class()
        mb = 1024 ** 2
        self.az = AppService(server=self.config["homeserver.address"],
                             domain=self.config["homeserver.domain"],
                             verify_ssl=self.config["homeserver.verify_ssl"],

                             as_token=self.config["appservice.as_token"],
                             hs_token=self.config["appservice.hs_token"],

                             tls_cert=self.config.get("appservice.tls_cert", None),
                             tls_key=self.config.get("appservice.tls_key", None),

                             bot_localpart=self.config["appservice.bot_username"],

                             log="mau.as",
                             loop=self.loop,

                             state_store=self.state_store,

                             real_user_content_key=self.real_user_content_key,

                             aiohttp_params={
                                 "client_max_size": self.config["appservice.max_body_size"] * mb
                             })

    def prepare_db(self) -> None:
        self.db = sql.create_engine(self.config["appservice.database"])
        Base.metadata.bind = self.db
        if not self.db.has_table("alembic_version"):
            self.log.critical("alembic_version table not found. "
                              "Did you forget to `alembic upgrade head`?")
            sys.exit(10)

    def prepare_bridge(self) -> None:
        self.matrix = self.matrix_class(az=self.az, config=self.config, loop=self.loop, bridge=self)

    async def start(self) -> None:
        self.log.debug("Starting appservice...")
        await self.az.start(self.config["appservice.hostname"], self.config["appservice.port"])
        await self.matrix.wait_for_connection()
        self.add_startup_actions(self.matrix.init_as_bot())
        await super().start()
        self.az.ready = True

    async def stop(self) -> None:
        await self.az.stop()
        await super().stop()



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Any, Optional
from abc import ABC
import logging
import asyncio

from mautrix.appservice import AppService
from mautrix.types import UserID


class BaseUser(ABC):
    log: logging.Logger = logging.getLogger("mau.user")
    az: AppService
    loop: asyncio.AbstractEventLoop

    is_whitelisted: bool
    mxid: UserID
    command_status: Optional[Dict[str, Any]]

    async def is_logged_in(self) -> bool:
        return False



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Tuple, Optional, Union, List, TYPE_CHECKING
from abc import ABC, abstractmethod
import logging
import asyncio
import time

from mautrix.types import (EventID, RoomID, UserID, Event, EventType, MessageEvent, MessageType,
                           MessageEventContent, StateEvent, Membership, MemberStateEventContent,
                           PresenceEvent, TypingEvent, ReceiptEvent, TextMessageEventContent,
                           EncryptedEvent)
from mautrix.errors import IntentError, MatrixError, MForbidden
from mautrix.appservice import AppService
from mautrix.util.logging import TraceLogger

from .commands import CommandProcessor

if TYPE_CHECKING:
    from .config import BaseBridgeConfig
    from .user import BaseUser
    from .portal import BasePortal
    from .puppet import BasePuppet
    from .bridge import Bridge

try:
    from .e2ee import EncryptionManager
except ImportError:
    EncryptionManager = None


class BaseMatrixHandler(ABC):
    log: TraceLogger = logging.getLogger("mau.mx")
    az: AppService
    commands: CommandProcessor
    config: 'BaseBridgeConfig'
    bridge: 'Bridge'
    e2ee: Optional[EncryptionManager]

    user_id_prefix: str
    user_id_suffix: str

    def __init__(self, az: AppService, config: 'BaseBridgeConfig',
                 command_processor: Optional[CommandProcessor] = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None,
                 bridge: Optional['Bridge'] = None) -> None:
        self.az = az
        self.config = config
        self.commands = command_processor or CommandProcessor(az=az, config=config, loop=loop,
                                                              bridge=bridge)
        self.az.matrix_event_handler(self.int_handle_event)

        self.e2ee = None
        if self.config["bridge.encryption.allow"]:
            if not EncryptionManager:
                self.log.error("Encryption enabled in config, but dependencies not installed.")
            elif not self.config["bridge.login_shared_secret"]:
                self.log.warning("Encryption enabled in config, but login_shared_secret not set.")
            else:
                self.e2ee = EncryptionManager(
                    bot_mxid=self.az.bot_mxid, device_name=bridge.name,
                    user_id_prefix=self.user_id_prefix, user_id_suffix=self.user_id_suffix,
                    login_shared_secret=self.config["bridge.login_shared_secret"],
                    homeserver_address=self.config["homeserver.address"], loop=loop)

    async def wait_for_connection(self) -> None:
        self.log.info("Ensuring connectivity to homeserver")
        errors = 0
        while True:
            try:
                await self.az.intent.whoami()
                break
            except MForbidden:
                raise
            except Exception:
                errors += 1
                if errors <= 6:
                    self.log.exception("Connection to homeserver failed, retrying in 10 seconds")
                    await asyncio.sleep(10)
                else:
                    raise

    async def init_as_bot(self) -> None:
        self.log.debug("Initializing appservice bot")
        displayname = self.config["appservice.bot_displayname"]
        if displayname:
            try:
                await self.az.intent.set_displayname(
                    displayname if displayname != "remove" else "")
            except Exception:
                self.log.exception("Failed to set bot displayname")

        avatar = self.config["appservice.bot_avatar"]
        if avatar:
            try:
                await self.az.intent.set_avatar_url(avatar if avatar != "remove" else "")
            except Exception:
                self.log.exception("Failed to set bot avatar")

        if self.e2ee:
            await self.e2ee.start()

    @abstractmethod
    async def get_user(self, user_id: UserID) -> 'BaseUser':
        pass

    @abstractmethod
    async def get_portal(self, room_id: RoomID) -> 'BasePortal':
        pass

    @abstractmethod
    async def get_puppet(self, user_id: UserID) -> 'BasePuppet':
        pass

    @staticmethod
    async def allow_message(user: 'BaseUser') -> bool:
        return user.is_whitelisted

    @staticmethod
    async def allow_command(user: 'BaseUser') -> bool:
        return user.is_whitelisted

    @staticmethod
    async def allow_bridging_message(user: 'BaseUser', portal: 'BasePortal') -> bool:
        return await user.is_logged_in()

    async def handle_leave(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        pass

    async def handle_kick(self, room_id: RoomID, user_id: UserID, kicked_by: UserID, reason: str,
                          event_id: EventID) -> None:
        pass

    async def handle_ban(self, room_id: RoomID, user_id: UserID, banned_by: UserID, reason: str,
                         event_id: EventID) -> None:
        pass

    async def handle_unban(self, room_id: RoomID, user_id: UserID, unbanned_by: UserID,
                           reason: str, event_id: EventID) -> None:
        pass

    async def handle_join(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        pass

    async def handle_member_info_change(self, room_id: RoomID, user_id: UserID,
                                        content: MemberStateEventContent,
                                        prev_content: MemberStateEventContent,
                                        event_id: EventID) -> None:
        pass

    async def handle_puppet_invite(self, room_id: RoomID, puppet: 'BasePuppet',
                                   invited_by: 'BaseUser', event_id: EventID) -> None:
        pass

    async def handle_invite(self, room_id: RoomID, user_id: UserID, inviter: 'BaseUser',
                            event_id: EventID) -> None:
        pass

    async def handle_reject(self, room_id: RoomID, user_id: UserID, reason: str, event_id: EventID
                            ) -> None:
        pass

    async def handle_disinvite(self, room_id: RoomID, user_id: UserID, disinvited_by: UserID,
                               reason: str, event_id: EventID) -> None:
        pass

    async def handle_event(self, evt: Event) -> None:
        """Called by :meth:`int_handle_event` for message events other than m.room.message."""

    async def handle_state_event(self, evt: StateEvent) -> None:
        """Called by :meth:`int_handle_event` for state events other than m.room.membership."""

    async def handle_ephemeral_event(self, evt: Union[ReceiptEvent, PresenceEvent, TypingEvent]
                                     ) -> None:
        pass

    async def send_permission_error(self, room_id: RoomID) -> None:
        await self.az.intent.send_notice(
            room_id,
            text="You are not whitelisted to use this bridge.\n\n"
                 "If you are the owner of this bridge, see the bridge.permissions "
                 "section in your config file.",
            html="<p>You are not whitelisted to use this bridge.</p>"
                 "<p>If you are the owner of this bridge, see the "
                 "<code>bridge.permissions</code> section in your config file.</p>")

    async def accept_bot_invite(self, room_id: RoomID, inviter: 'BaseUser') -> None:
        tries = 0
        while tries < 5:
            try:
                await self.az.intent.join_room(room_id)
                break
            except (IntentError, MatrixError):
                tries += 1
                wait_for_seconds = (tries + 1) * 10
                if tries < 5:
                    self.log.exception(f"Failed to join room {room_id} with bridge bot, "
                                       f"retrying in {wait_for_seconds} seconds...")
                    await asyncio.sleep(wait_for_seconds)
                else:
                    self.log.exception(f"Failed to join room {room_id}, giving up.")
                    return

        if not await self.allow_command(inviter):
            await self.send_permission_error(room_id)
            await self.az.intent.leave_room(room_id)
            return

        await self.send_welcome_message(room_id, inviter)

    async def send_welcome_message(self, room_id: RoomID, inviter: 'BaseUser') -> None:
        await self.az.intent.send_notice(
            room_id=room_id,
            text=f"Hello, I'm a bridge bot. Use `{self.commands.command_prefix} help` for help.",
            html=f"Hello, I'm a bridge bot. Use <code>{self.commands.command_prefix} help</code> "
                 f"for help.")

    async def int_handle_invite(self, room_id: RoomID, user_id: UserID, invited_by: UserID,
                                event_id: EventID) -> None:
        self.log.debug(f"{invited_by} invited {user_id} to {room_id}")
        inviter = await self.get_user(invited_by)
        if inviter is None:
            self.log.exception(f"Failed to find user with Matrix ID {invited_by}")
            return
        elif user_id == self.az.bot_mxid:
            await self.accept_bot_invite(room_id, inviter)
            return
        elif not await self.allow_command(inviter):
            return

        puppet = await self.get_puppet(user_id)
        if puppet:
            await self.handle_puppet_invite(room_id, puppet, inviter, event_id)
            return

        await self.handle_invite(room_id, user_id, inviter, event_id)

    def is_command(self, message: MessageEventContent) -> Tuple[bool, str]:
        text = message.body
        prefix = self.config["bridge.command_prefix"]
        is_command = text.startswith(prefix)
        if is_command:
            text = text[len(prefix) + 1:].lstrip()
        return is_command, text

    async def handle_message(self, room_id: RoomID, user_id: UserID, message: MessageEventContent,
                             event_id: EventID) -> None:
        sender = await self.get_user(user_id)
        if not sender or not await self.allow_message(sender):
            self.log.debug(f"Ignoring message {event_id} from {user_id} to {room_id}:"
                           " User is not whitelisted.")
            return
        self.log.debug(f"Received Matrix event {event_id} from {sender.mxid} in {room_id}")
        self.log.trace("Event %s content: %s", event_id, message)

        if isinstance(message, TextMessageEventContent):
            message.trim_reply_fallback()

        is_command, text = self.is_command(message)
        portal = await self.get_portal(room_id)
        if not is_command and portal and await self.allow_bridging_message(sender, portal):
            await portal.handle_matrix_message(sender, message, event_id)
            return

        if message.msgtype != MessageType.TEXT or not await self.allow_command(sender):
            return

        is_management = await self.is_management(room_id)

        if is_command or is_management:
            try:
                command, arguments = text.split(" ", 1)
                args = arguments.split(" ")
            except ValueError:
                # Not enough values to unpack, i.e. no arguments
                command = text
                args = []
            await self.commands.handle(room_id, event_id, sender, command, args, message,
                                       is_management, is_portal=portal is not None)

    async def is_management(self, room_id: RoomID) -> bool:
        try:
            members = await self.az.intent.get_room_members(room_id)
            return len(members) == 2 and self.az.bot_mxid in members
        except MatrixError:
            return False

    def filter_matrix_event(self, evt: Event) -> bool:
        if not isinstance(evt, (MessageEvent, StateEvent)):
            return False
        return evt.sender == self.az.bot_mxid

    async def try_handle_sync_event(self, evt: Event) -> None:
        try:
            if isinstance(evt, (ReceiptEvent, PresenceEvent, TypingEvent)):
                await self.handle_ephemeral_event(evt)
            else:
                self.log.trace("Unknown event type received from sync: %s", evt)
        except Exception:
            self.log.exception("Error handling manually received Matrix event")

    async def handle_encrypted(self, evt: EncryptedEvent) -> None:
        await self.int_handle_event(self.e2ee.decrypt(evt))

    async def enable_dm_encryption(self, portal: 'BasePortal', members: List[UserID]) -> bool:
        try:
            await portal.main_intent.invite_user(portal.mxid, self.az.bot_mxid)
            await self.az.intent.join_room_by_id(portal.mxid)
            await portal.main_intent.send_state_event(portal.mxid, EventType.ROOM_ENCRYPTION, {
                "algorithm": "m.megolm.v1.aes-sha2"
            })
        except Exception:
            self.log.warning(f"Failed to enable end-to-bridge encryption in {portal.mxid}",
                             exc_info=True)
            return False

        portal.encrypted = True
        await self.e2ee.add_room(portal.mxid, members=members + [self.az.bot_mxid], encrypted=True)
        return True

    async def int_handle_event(self, evt: Event) -> None:
        if isinstance(evt, StateEvent) and evt.type == EventType.ROOM_MEMBER and self.e2ee:
            await self.e2ee.handle_room_membership(evt)
        if self.filter_matrix_event(evt):
            return
        self.log.trace("Received event: %s", evt)
        start_time = time.time()

        if evt.type == EventType.ROOM_MEMBER:
            evt: StateEvent
            prev_content = evt.unsigned.prev_content or MemberStateEventContent()
            prev_membership = prev_content.membership if prev_content else Membership.JOIN
            if evt.content.membership == Membership.INVITE:
                await self.int_handle_invite(evt.room_id, UserID(evt.state_key), evt.sender,
                                             evt.event_id)
            elif evt.content.membership == Membership.LEAVE:
                if prev_membership == Membership.BAN:
                    await self.handle_unban(evt.room_id, UserID(evt.state_key), evt.sender,
                                            evt.content.reason, evt.event_id)
                elif prev_membership == Membership.INVITE:
                    if evt.sender == evt.state_key:
                        await self.handle_reject(evt.room_id, UserID(evt.state_key),
                                                 evt.content.reason, evt.event_id)
                    else:
                        await self.handle_disinvite(evt.room_id, UserID(evt.state_key), evt.sender,
                                                    evt.content.reason, evt.event_id)
                elif evt.sender == evt.state_key:
                    await self.handle_leave(evt.room_id, UserID(evt.state_key), evt.event_id)
                else:
                    await self.handle_kick(evt.room_id, UserID(evt.state_key), evt.sender,
                                           evt.content.reason, evt.event_id)
            elif evt.content.membership == Membership.BAN:
                await self.handle_ban(evt.room_id, UserID(evt.state_key), evt.sender,
                                      evt.content.reason, evt.event_id)
            elif evt.content.membership == Membership.JOIN:
                if prev_membership != Membership.JOIN:
                    await self.handle_join(evt.room_id, UserID(evt.state_key), evt.event_id)
                else:
                    await self.handle_member_info_change(evt.room_id, UserID(evt.state_key),
                                                         evt.content, prev_content, evt.event_id)
        elif evt.type in (EventType.ROOM_MESSAGE, EventType.STICKER):
            evt: MessageEvent
            if evt.type != EventType.ROOM_MESSAGE:
                evt.content.msgtype = MessageType(str(evt.type))
            await self.handle_message(evt.room_id, evt.sender, evt.content, evt.event_id)
        elif evt.type == EventType.ROOM_ENCRYPTION and self.e2ee:
            await self.e2ee.handle_room_encryption(evt)
        elif evt.type == EventType.ROOM_ENCRYPTED and self.e2ee:
            await self.handle_encrypted(evt)
        else:
            if evt.type.is_state and isinstance(evt, StateEvent):
                await self.handle_state_event(evt)
            else:
                await self.handle_event(evt)

        await self.log_event_handle_duration(evt, time.time() - start_time)

    async def log_event_handle_duration(self, evt: Event, duration: float) -> None:
        pass



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional
import json

from sqlalchemy import Column, String, types

from mautrix.types import RoomID, PowerLevelStateEventContent
from mautrix.util.db import Base


class PowerLevelType(types.TypeDecorator):
    impl = types.Text

    @property
    def python_type(self):
        return PowerLevelStateEventContent

    def process_bind_param(self, value: PowerLevelStateEventContent, dialect) -> Optional[str]:
        if value is not None:
            return json.dumps(value.serialize())
        return None

    def process_result_value(self, value: str, dialect) -> Optional[PowerLevelStateEventContent]:
        if value is not None:
            return PowerLevelStateEventContent.deserialize(json.loads(value))
        return None

    def process_literal_param(self, value, dialect):
        return value


class RoomState(Base):
    __tablename__ = "mx_room_state"

    room_id: RoomID = Column(String(255), primary_key=True)
    power_levels: PowerLevelStateEventContent = Column(PowerLevelType, nullable=True)

    @property
    def has_power_levels(self) -> bool:
        return bool(self.power_levels)

    @classmethod
    def get(cls, room_id: RoomID) -> Optional['RoomState']:
        return cls._select_one_or_none(cls.c.room_id == room_id)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, Callable, Union, Optional, Awaitable, Any, Type, TYPE_CHECKING
from enum import Enum, Flag, auto
from time import time
import asyncio

from ..errors import MUnknownToken
from ..api import JSON
from .api.types import (EventType, MessageEvent, StateEvent, StrippedStateEvent, Event, FilterID,
                        Filter, PresenceState)
from .api import ClientAPI
from .store import ClientStore, MemoryClientStore

if TYPE_CHECKING:
    from .dispatcher import Dispatcher

EventHandler = Callable[[Event], Awaitable[None]]


class SyncStream(Flag):
    INTERNAL = auto()

    JOINED_ROOM = auto()
    INVITED_ROOM = auto()
    LEFT_ROOM = auto()

    TIMELINE = auto()
    STATE = auto()
    EPHEMERAL = auto()
    ACCOUNT_DATA = auto()


class InternalEventType(Enum):
    SYNC_STARTED = auto()
    SYNC_ERRORED = auto()
    SYNC_SUCCESSFUL = auto()
    SYNC_STOPPED = auto()

    JOIN = auto()
    PROFILE_CHANGE = auto()
    INVITE = auto()
    REJECT_INVITE = auto()
    DISINVITE = auto()
    LEAVE = auto()
    KICK = auto()
    BAN = auto()
    UNBAN = auto()


class Client(ClientAPI):
    """Client is a high-level wrapper around the client API."""
    store: ClientStore
    global_event_handlers: List[EventHandler]
    event_handlers: Dict[Union[EventType, InternalEventType], List[EventHandler]]
    dispatchers: Dict[Type['Dispatcher'], 'Dispatcher']
    syncing_task: Optional[asyncio.Future]
    ignore_initial_sync: bool
    ignore_first_sync: bool
    presence: PresenceState

    def __init__(self, *args, store: ClientStore = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.store = store or MemoryClientStore()
        self.global_event_handlers = []
        self.event_handlers = {}
        self.dispatchers = {}
        self.syncing_task = None
        self.ignore_initial_sync = False
        self.ignore_first_sync = False
        self.presence = PresenceState.ONLINE

    def on(self, var: Union[EventHandler, EventType, InternalEventType]
           ) -> Union[EventHandler, Callable[[EventHandler], EventHandler]]:
        """
        Add a new event handler. This method is for decorator usage.
        Use :meth:`add_event_handler` if you don't use a decorator.

        Args:
            var: Either the handler function or the event type to handle.

        Returns:
            If ``var`` was the handler function, the handler function is returned.

            If ``var`` was an event type, a function that takes the handler function as an argument
            is returned.

        Examples:
            >>> client = Client(...)
            >>> @client.on(EventType.ROOM_MESSAGE)
            >>> def handler(event: MessageEvent) -> None:
            ...     pass
        """
        if isinstance(var, (EventType, InternalEventType)):
            def decorator(func: EventHandler) -> EventHandler:
                self.add_event_handler(var, func)
                return func

            return decorator
        else:
            self.add_event_handler(EventType.ALL, var)
            return var

    def add_dispatcher(self, dispatcher_type: Type['Dispatcher']) -> None:
        if dispatcher_type in self.dispatchers:
            return
        self.dispatchers[dispatcher_type] = dispatcher_type(self)
        self.dispatchers[dispatcher_type].register()

    def remove_dispatcher(self, dispatcher_type: Type['Dispatcher']) -> None:
        if dispatcher_type not in self.dispatchers:
            return
        self.dispatchers[dispatcher_type].unregister()
        del self.dispatchers[dispatcher_type]

    def add_event_handler(self, event_type: Union[InternalEventType, EventType],
                          handler: EventHandler) -> None:
        """
        Add a new event handler.

        Args:
            event_type: The event type to add. If not specified, the handler will be called for all
                event types.
            handler: The handler function to add.
        """
        if not isinstance(event_type, (EventType, InternalEventType)):
            raise ValueError("Invalid event type")
        if event_type == EventType.ALL:
            self.global_event_handlers.append(handler)
        else:
            self.event_handlers.setdefault(event_type, []).append(handler)

    def remove_event_handler(self, event_type: Union[EventType, InternalEventType],
                             handler: EventHandler) -> None:
        """
        Remove an event handler.

        Args:
            handler: The handler function to remove.
            event_type: The event type to remove the handler function from.
        """
        if not isinstance(event_type, (EventType, InternalEventType)):
            raise ValueError("Invalid event type")
        try:
            if event_type == EventType.ALL:
                self.global_event_handlers.remove(handler)
            else:
                handlers = self.event_handlers[event_type]
                handlers.remove(handler)
                if len(handlers) == 0:
                    del self.event_handlers[event_type]
        except (KeyError, ValueError):
            pass

    async def dispatch_event(self, event: Event, source: SyncStream) -> None:
        """
        Send the given event to all applicable event handlers.

        Args:
            event: The event to send.
            source: The sync stream the event was received in.
        """
        if isinstance(event, MessageEvent):
            event.content.trim_reply_fallback()
        if event.type.is_state and event.state_key is None:
            self.log.debug(f"Not sending {event.event_id} to handlers: expected state_key.")
            return
        elif event.type.is_account_data and not source & SyncStream.ACCOUNT_DATA:
            self.log.debug(f"Not sending {event.event_id} to handlers: got account_data event "
                           f"type in non-account_data sync stream.")
            return
        elif event.type.is_ephemeral and not source & SyncStream.EPHEMERAL:
            self.log.debug(f"Not sending {event.event_id} to handlers: got ephemeral event "
                           f"type in non-ephemeral sync stream.")
            return

        setattr(event, "source", source)
        await self.dispatch_manual_event(event.type, event, include_global_handlers=True)

    async def dispatch_manual_event(self, event_type: Union[EventType, InternalEventType],
                                    data: Any, include_global_handlers: bool = False) -> None:
        handlers = self.event_handlers.get(event_type, [])
        if include_global_handlers:
            handlers = self.global_event_handlers + handlers
        for handler in handlers:
            try:
                await handler(data)
            except Exception:
                self.log.exception("Failed to run handler")

    def dispatch_internal_event(self, event_type: InternalEventType, **kwargs: Any
                                ) -> Awaitable[None]:
        kwargs["source"] = SyncStream.INTERNAL
        return self.dispatch_manual_event(event_type, kwargs,
                                          include_global_handlers=False)

    def handle_sync(self, data: JSON) -> None:
        """
        Handle a /sync object.

        Args:
            data: The data from a /sync request.
        """
        rooms = data.get("rooms", {})
        for room_id, room_data in rooms.get("join", {}).items():
            for raw_event in room_data.get("state", {}).get("events", []):
                raw_event["room_id"] = room_id
                asyncio.ensure_future(
                    self.dispatch_event(StateEvent.deserialize(raw_event),
                                        source=SyncStream.JOINED_ROOM | SyncStream.STATE),
                    loop=self.loop)

            for raw_event in room_data.get("timeline", {}).get("events", []):
                raw_event["room_id"] = room_id
                asyncio.ensure_future(
                    self.dispatch_event(Event.deserialize(raw_event),
                                        source=SyncStream.JOINED_ROOM | SyncStream.TIMELINE),
                    loop=self.loop)
        for room_id, room_data in rooms.get("invite", {}).items():
            events: List[Dict[str, Any]] = room_data.get("invite_state", {}).get("events", [])
            for raw_event in events:
                raw_event["room_id"] = room_id
            raw_invite = next(raw_event for raw_event in events
                              if raw_event.get("type", "") == "m.room.member"
                              and raw_event.get("state_key", "") == self.mxid)
            # These aren't required by the spec, so make sure they're set
            raw_invite.setdefault("event_id", None)
            raw_invite.setdefault("origin_server_ts", int(time() * 1000))

            invite = StateEvent.deserialize(raw_invite)
            invite.unsigned.invite_room_state = [StrippedStateEvent.deserialize(raw_event)
                                                 for raw_event in events
                                                 if raw_event != raw_invite]
            asyncio.ensure_future(
                self.dispatch_event(invite, source=SyncStream.INVITED_ROOM | SyncStream.STATE),
                loop=self.loop)
        for room_id, room_data in rooms.get("leave", {}).items():
            for raw_event in room_data.get("timeline", {}).get("events", []):
                if "state_key" in raw_event:
                    raw_event["room_id"] = room_id
                    asyncio.ensure_future(
                        self.dispatch_event(StateEvent.deserialize(raw_event),
                                            source=SyncStream.LEFT_ROOM | SyncStream.TIMELINE),
                        loop=self.loop)

    def start(self, filter_data: Optional[Union[FilterID, Filter]]) -> asyncio.Future:
        """
        Start syncing with the server. Can be stopped with :meth:`stop`.

        Args:
            filter_data: The filter data or filter ID to use for syncing.
        """
        if self.syncing_task is not None:
            self.syncing_task.cancel()
        self.syncing_task = asyncio.ensure_future(self._try_start(filter_data), loop=self.loop)
        return self.syncing_task

    async def _try_start(self, filter_data: Optional[Union[FilterID, Filter]]) -> None:
        try:
            if isinstance(filter_data, Filter):
                filter_data = await self.create_filter(filter_data)
            await self._start(filter_data)
        except asyncio.CancelledError:
            self.log.debug("Syncing cancelled")
        except Exception as e:
            self.log.exception("Fatal error while syncing")
            await self.dispatch_internal_event(InternalEventType.SYNC_STOPPED, error=e)
            return
        else:
            self.log.debug("Syncing stopped")
        await self.dispatch_internal_event(InternalEventType.SYNC_STOPPED, error=None)

    async def _start(self, filter_id: Optional[FilterID]) -> None:
        fail_sleep = 5
        is_first = True

        self.log.debug("Starting syncing")
        await self.dispatch_internal_event(InternalEventType.SYNC_STARTED)
        while True:
            try:
                data = await self.sync(since=self.store.next_batch, filter_id=filter_id,
                                       set_presence=self.presence)
                fail_sleep = 5
            except (asyncio.CancelledError, MUnknownToken):
                raise
            except Exception as e:
                self.log.exception(f"Sync request errored, waiting {fail_sleep}"
                                   " seconds before continuing")
                await self.dispatch_internal_event(InternalEventType.SYNC_ERRORED, error=e,
                                                   sleep_for=fail_sleep)
                await asyncio.sleep(fail_sleep, loop=self.loop)
                if fail_sleep < 320:
                    fail_sleep *= 2
                continue

            is_initial = not self.store.next_batch
            data["net.maunium.mautrix"] = {
                "is_initial": is_initial,
                "is_first": is_first,
            }
            self.store.next_batch = data.get("next_batch")
            await self.dispatch_internal_event(InternalEventType.SYNC_SUCCESSFUL, data=data)
            if (self.ignore_first_sync and is_first) or (self.ignore_initial_sync and is_initial):
                is_first = False
                continue
            is_first = False
            try:
                self.handle_sync(data)
            except Exception:
                self.log.exception("Sync handling errored")

    def stop(self) -> None:
        """
        Stop a sync started with :meth:`start`.
        """
        if self.syncing_task:
            self.syncing_task.cancel()
            self.syncing_task = None
